/* 
 * Copyright (C) 2008-2009 by Emmanuel Azencot under the GNU LGPL
 * license version 2.0 or 2.1.  You should have received a copy of the
 * LGPL license along with this library if you did not you can find it
 * at http://www.gnu.org/.
 */
/*
 * Azencot : Sat Dec  6 02:31:40 CET 2008
 *  Creation
 */

/*
 gcc -g -I. -I./excp -o ring_alloc ./ring.c ring_alloc.c
 */
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "config.h"
#include "ring.h"
#include "ring_alloc.h"

#if defined(RING_ALLOC_BLK_CHK) || defined(RING_ALLOC_POOL_CHK)
#include <assert.h>
#endif

#ifdef RING_ALLOC_POOL_CHK
  #undef RING_ALLOC_POOL_CHK
  #define RING_ALLOC_POOL_CHK  0x37DB5104
#endif
#ifdef RING_ALLOC_BLK_CHK
  #undef RING_ALLOC_BLK_CHK
  #define RING_ALLOC_BLK_CHK  0x37DB5104
#endif

struct s_blk {
#ifdef RING_ALLOC_BLK_CHK
   unsigned long chk_1[2];
#endif
   struct s_ring brother;
   size_t size;
#ifdef RING_ALLOC_BLK_CHK
   unsigned long chk_2[2];
#endif
};
struct s_mempool {
#ifdef RING_ALLOC_POOL_CHK
   unsigned long chk_1;
#endif
   size_t size;
#ifdef RING_ALLOC_POOL_CHK
   unsigned long chk_2;
#endif
#ifdef RING_ALLOC_STATS
   struct s_mempool_stats stats;
#endif
#ifdef RING_ALLOC_POOL_CHK
   unsigned long chk_3;
#endif
   struct s_blk *alloc;
#ifdef RING_ALLOC_POOL_CHK
   unsigned long chk_4;
#endif
   struct s_blk *free;
#ifdef RING_ALLOC_POOL_CHK
   unsigned long chk_5;
#endif
   struct s_blk mem[0];
};
#define rotate(val32, bits) ( ((unsigned long)(val32))<<(bits) | ((unsigned long)(val32))>>(32-(bits)))

#ifdef RING_ALLOC_BLK_CHK
#define m_blk_setup(blk) { \
   blk->chk_1[0] = rotate(RING_ALLOC_BLK_CHK,1) ; \
   blk->chk_1[1] = rotate(RING_ALLOC_BLK_CHK,2) ; \
   blk->chk_2[0] = rotate(RING_ALLOC_BLK_CHK,3) ; \
   blk->chk_2[1] = rotate(RING_ALLOC_BLK_CHK,4) ; \
}
#include <stdio.h>
int f_mempool_check_blk(struct s_mempool *pool, struct s_blk *list, struct s_blk *blk) {
   struct s_blk *mem;
   int neibourgs = 0;
   int saved_errno = errno;

   assert( blk );
   assert( pool->mem <= blk );
   assert( ((char *)blk + sizeof(struct s_blk) + blk->size) < ((char *)pool->mem + pool->size) );
   assert ( list );
   assert( blk->chk_1[0] == rotate(RING_ALLOC_BLK_CHK,1) );
   assert( blk->chk_1[1] == rotate(RING_ALLOC_BLK_CHK,2) );
   assert( blk->chk_2[0] == rotate(RING_ALLOC_BLK_CHK,3) );
   assert( blk->chk_2[1] == rotate(RING_ALLOC_BLK_CHK,4) );
   /* can use it cause brother is the 1st field */
   assert ( m_ring_is_in(blk, list, brother) );

   // assert ( (m_ring_is_in(blk, pool->free, brother) ^ m_ring_is_in(blk, pool->alloc, brother)) );
   neibourgs += m_ring_is_in(blk, pool->alloc, brother);
   neibourgs += 2*m_ring_is_in(blk, pool->free, brother);
   assert ( neibourgs == 2 || neibourgs == 1 );
   neibourgs = 0;

//   printf("\nblk   (%08x <-> %08x)\n",  blk,  ((char *)blk + blk->size + sizeof(struct s_blk)) );
//   printf("pool  (%08x <-> %08x)\n",  pool->mem,  ((char *)pool->mem + pool->size - sizeof(struct s_mempool)) );
   for (mem = pool->alloc; mem; mem = m_ring_list(pool->alloc, mem, brother)) {
//      printf("alloc (%08x <-> %08x)\n",  mem,  ((char *)mem + mem->size + sizeof(struct s_blk)) );
      /* bloc is behind this bloc ? */
      if ( ((char *)mem + mem->size + sizeof(struct s_blk)) == (char *)blk ) neibourgs |= 1;
       /* bloc is after the end of this bloc ? */
      if ( ((char *)blk + blk->size + sizeof(struct s_blk)) == (char *)mem ) neibourgs |= 2;
   }
   for (mem = pool->free; mem; mem = m_ring_list(pool->free, mem, brother) ) {
//      printf("free  (%08x <-> %08x)\n",  mem,  ((char *)mem + mem->size + sizeof(struct s_blk)) );
      /* bloc is behind this bloc ? */
      if ( ((char *)mem + mem->size + sizeof(struct s_blk)) == (char *)blk ) neibourgs |= 1;
      /* bloc is after the end of this bloc ? */
      if ( ((char *)blk + blk->size + sizeof(struct s_blk)) == (char *)mem ) neibourgs |= 2;
   }
//   printf("neibourgs = %d, ", neibourgs); 
   if ( blk == pool->mem ) neibourgs |= 1;
   if ( ((char *)blk + blk->size + sizeof(struct s_blk)) == ((char *)pool->mem + pool->size - sizeof(struct s_mempool)) ) neibourgs |= 2;
//   printf("neibourgs = %d\n", neibourgs); fflush(stdout);
   assert( neibourgs == 3 );
   errno = saved_errno;
   return 1;
}
#else
#define f_mempool_check_blk(list, blk) (1)
#define m_blk_setup(blk) ;
#endif

#ifdef RING_ALLOC_POOL_CHK
int f_mempool_check(struct s_mempool *mempool) {
   size_t o_count_a = 0, b_count_a = 0;
   size_t o_count_f = 0, b_count_f = 0;
   struct s_blk *blk = mempool->free;
   
   assert( mempool );
   assert( mempool->chk_1 == rotate(RING_ALLOC_POOL_CHK,1) );
   assert( mempool->chk_2 == rotate(RING_ALLOC_POOL_CHK,2) );
   assert( mempool->chk_3 == rotate(RING_ALLOC_POOL_CHK,3) );
   assert( mempool->chk_4 == rotate(RING_ALLOC_POOL_CHK,4) );
   assert( mempool->chk_5 == rotate(RING_ALLOC_POOL_CHK,5) );
   assert( mempool->size != 0 && mempool->size < 0x80000000 );
#ifdef RING_ALLOC_STATS
   assert( mempool->stats.instant.bytes.alloc <= mempool->stats.cumul.bytes.alloc );
   assert( mempool->stats.instant.blocks.alloc <= mempool->stats.cumul.blocks.alloc );
   assert( (sizeof(struct s_blk) * (
               mempool->stats.instant.blocks.free + 
               mempool->stats.instant.blocks.alloc) +
            mempool->stats.instant.bytes.free +
            mempool->stats.instant.bytes.alloc)
            == (mempool->size - sizeof(struct s_mempool)) );
#endif
   for (blk = mempool->free; blk; blk = m_ring_list(mempool->free, blk, brother)) {
       assert( f_mempool_check_blk(mempool, mempool->free, blk) );
       o_count_f += blk->size;
       ++b_count_f;
   }
#ifdef RING_ALLOC_STATS
   assert ( o_count_f == mempool->stats.instant.bytes.free );
   assert ( b_count_f == mempool->stats.instant.blocks.free );
#endif
   for (blk = mempool->alloc; blk; blk = m_ring_list(mempool->alloc, blk, brother)) {
       assert( f_mempool_check_blk(mempool, mempool->alloc, blk) );
       o_count_a += blk->size;
       ++b_count_a;
   }
#ifdef RING_ALLOC_STATS
   assert ( o_count_a == mempool->stats.instant.bytes.alloc );
   assert ( b_count_a == mempool->stats.instant.blocks.alloc );
#endif
   assert( (sizeof(struct s_blk) * (b_count_a + b_count_f) + o_count_a + o_count_f)
           == (mempool->size - sizeof(struct s_mempool)) );
   return 1;
}
#else
#define f_mempool_check(mempool) (1)
#endif

to_mempool f_mempool_init(size_t size, void *start) {
   struct s_mempool *pool = (typeof(pool))(start);
   if ( !start ) { errno = EFAULT; return 0; }
   if ( size < sizeof(struct s_mempool) + sizeof(struct s_blk)) { errno = ENOMEM; return 0; }
   memset(pool, 0, sizeof(struct s_mempool));
#ifdef RING_ALLOC_POOL_CHK
   pool->chk_1 = rotate(RING_ALLOC_POOL_CHK,1);
   pool->chk_2 = rotate(RING_ALLOC_POOL_CHK,2);
   pool->chk_3 = rotate(RING_ALLOC_POOL_CHK,3);
   pool->chk_4 = rotate(RING_ALLOC_POOL_CHK,4);
   pool->chk_5 = rotate(RING_ALLOC_POOL_CHK,5);
#endif
   pool->size = size;
   pool->alloc = 0;
   pool->free = 0;
   pool->free = m_ring_link(pool->free, brother, pool->mem);
   pool->free->size = pool->size -sizeof(struct s_mempool) -sizeof(struct s_blk) ;
   m_blk_setup(pool->free);
#ifdef RING_ALLOC_STATS
   pool->stats.instant.bytes.free = pool->free->size;
   pool->stats.instant.blocks.free = 1;
#endif
   return pool;
}
int f_mempool_info(struct s_mempool_info *info) {
   if ( !info ) { errno = EFAULT; return -1; }
   info->version = RING_ALLOC_VERSION;
   info->release = RING_ALLOC_RELEASE;
   info->options.stats = 0; 
   info->options.chk_pool = 0;
   info->options.chk_blk = 0;
#ifdef RING_ALLOC_STATS
   info->options.stats = 1; 
#endif
#ifdef RING_ALLOC_POOL_CHK
   info->options.chk_pool = 1;
#endif
#ifdef RING_ALLOC_BLK_CHK
   info->options.chk_blk = 1;
#endif
   info->pool_head_size = sizeof(struct s_mempool); /* gcc bug ? */
   info->block_head_size = sizeof(struct s_blk);
   return 0;
}

#ifdef RING_ALLOC_STATS
/* stats space is provided by caller */
int f_mempool_stats(to_mempool mempool, struct s_mempool_stats *stats) {
   f_mempool_check(mempool);
   if ( !stats ) { errno = EFAULT; return -1; }
   *stats = mempool->stats;
   return 0;
}
/* increment stats alloc counters */
#define m_mempool_stats_alloc(mempool, size, split) { \
   ++mempool->stats.instant.blocks.alloc; \
   mempool->stats.instant.bytes.alloc += (size); \
   mempool->stats.instant.bytes.free -= (size); \
 \
   ++mempool->stats.cumul.blocks.alloc; \
   mempool->stats.cumul.bytes.alloc += (size); \
 \
   if ( !(split) ) { \
      --mempool->stats.instant.blocks.free; \
   } else { \
      mempool->stats.instant.bytes.free -= sizeof(struct s_blk); \
   }  \
}
/* decrement stats alloc counters */
#define m_mempool_stats_free(mempool, size, nb_cat) { \
   --mempool->stats.instant.blocks.alloc; \
   mempool->stats.instant.bytes.free += (size) + (nb_cat)*sizeof(struct s_blk); \
   mempool->stats.instant.bytes.alloc -= (size); \
 \
   ++mempool->stats.cumul.blocks.free; \
   mempool->stats.cumul.bytes.free += (size); \
 \
  mempool->stats.instant.blocks.free += (1 -(nb_cat)); \
}
#else /* RING_ALLOC_STATS */
/* stats not available */
int f_mempool_stats(to_mempool mempool, struct s_mempool_stats *stats) {
   memset(stats, 0, sizeof(*stats));
   errno = ENOTSUP; return -1;
}
/* internal use */
#define m_mempool_stats_alloc(mempool, size, split) ;
#define m_mempool_stats_free(mempool, size, nb_cat) ;
#endif /* RING_ALLOC_STATS */

void *f_mempool_malloc(to_mempool mempool, size_t size) {
   int split = 0; /* stat */
   struct s_blk *blk = mempool->free, *best = 0;
   /* can help, disable/enable at compile */
   f_mempool_check(mempool);
   if ( !blk ) { errno = ENOMEM; return 0; }

   /* find smallest free block */
   for (blk = mempool->free; blk; blk = m_ring_list(mempool->free, blk, brother)) {
      if ( blk->size < size ) continue;
      if ( best && best->size < blk->size) continue;
      best = blk;
   }
   /* nothing bigger enougth */
   if ( !best ) { errno = ENOMEM; return 0; }
   /* if best is bigger enougth to contain an other bloc, split it */
   if ( sizeof(struct s_blk) < best->size - size ) {
      split = 1; /* stat */
      /* begin stays in free list, reduce size */
      best->size -= (sizeof(struct s_blk) + size);
      /* move to begin of new block */
      best = (struct s_blk *)((char *)best + sizeof(struct s_blk) + best->size);
      /* put new block in allocated */
      best->size = size;
      m_blk_setup(best);
      mempool->alloc = m_ring_link(mempool->alloc, brother, best);
   } else {
      /* it fits well enougth, just move it to allocated blocks */
      mempool->free = m_ring_unlink(best, brother);
      mempool->alloc = m_ring_link(mempool->alloc, brother, best);
   }
   m_mempool_stats_alloc(mempool, best->size, split);
   f_mempool_check(mempool);
   return best+1;
}
int f_mempool_free(to_mempool mempool, void *mem) {
   int cat = 0, size;
   struct s_blk *blk = mempool->free;
   struct s_blk *umem = (typeof(blk))mem -1;

   f_mempool_check(mempool);
   if ( !mem ) return 0;

   if( mempool->mem > umem || 
      ((char *)mempool->mem + mempool->size) <= ((char *)umem + sizeof(struct s_blk) + umem->size) ) {
      errno = ENOTDIR;
      return -1;
   }

   /* can use it cause brother is the 1st field */
   if ( !m_ring_is_in(umem, mempool->alloc, brother) ) {
      errno = ENOENT; return -1;
   }
   size = umem->size;
   mempool->alloc = m_ring_unlink(umem, brother);
   /* find if this free bloc can be collapsed to a neibourg */
   for (blk = mempool->free; blk; blk = m_ring_list(mempool->free, blk, brother)) {
      /* freed bloc is behind this bloc ? */
      if ( ((char *)umem + umem->size + sizeof(struct s_blk)) == (char *)blk ) {
         mempool->free = m_ring_unlink(blk, brother);
         umem->size += blk->size + sizeof(struct s_blk);
         ++cat;
         break;
      }
   }
   for (blk = mempool->free; blk; blk = m_ring_list(mempool->free, blk, brother)) {
      /* freed bloc is at the end of this bloc ? */
      if ( ((char *)blk + blk->size + sizeof(struct s_blk)) == (char *)umem ) {
         blk->size += umem->size + sizeof(struct s_blk);
         ++cat;
         goto stats;
      }
   }
   mempool->free = m_ring_link(mempool->free, brother, umem);
stats:
   m_mempool_stats_free(mempool, size, cat);
   f_mempool_check(mempool);
   return 0;
}
void *f_mempool_realloc(to_mempool mempool, void *mem, size_t size) {
   int split = 0;
   struct s_blk *blk = mempool->free;
   struct s_blk *umem = (typeof(blk))mem -1;

   if ( !mem ) return f_mempool_malloc(mempool, size);
   if( mempool->mem > umem || 
      ((char *)mempool->mem + mempool->size) <= ((char *)umem + sizeof(struct s_blk) + umem->size) ) {
      errno = ENOTDIR;
      return 0;
   }

   if ( size < umem->size ) return mem;
   /* need blok after */
   for (blk = mempool->free; blk; blk = m_ring_list(mempool->free, blk, brother)) {
      if ( ((char *)umem + umem->size + sizeof(struct s_blk)) == (char *)blk ) {
         mempool->free = m_ring_unlink(blk, brother);
         umem->size += blk->size + sizeof(struct s_blk);

         if ( sizeof(struct s_blk) < umem->size - size ) {
            split = 1;
            /* begin stays in free list, reduce size */
            umem->size -= sizeof(struct s_blk) + size;
            /* move to begin of new block */
            umem = (struct s_blk *)((char *)umem + sizeof(struct s_blk) + size);
            /* put new block in allocated */
            umem->size = size;
            mempool->alloc = m_ring_link(mempool->alloc, brother, umem);
         } else {
            /* it fits well enougth, just move it to allocated blocks */
            mempool->free = m_ring_unlink(umem, brother);
            mempool->alloc = m_ring_link(mempool->alloc, brother, umem);
         }
         return umem -1;
      }
   }
   if ( !(umem = f_mempool_malloc(mempool, size)) ) return 0;
   memcpy(umem, mem, size);
   if ( f_mempool_free(mempool, mem) < 0 ) return 0;
   return umem -1;
}

void *f_mempool_calloc(to_mempool mempool, size_t nmemb, size_t size) {
   void *mem;

   if ( !(mem = f_mempool_malloc(mempool, nmemb*size)) == 0) return 0;
   memset(mem, 0, nmemb*size);

   return mem;
}

