/* 
 * Copyright (C) 2008-2009 by Emmanuel Azencot under the GNU GPL
 * license version 2.0 or 2.1.  You should have received a copy of the
 * GPL license along with this library if you did not you can find it
 * at http://www.gnu.org/.
 */
/*
 * Azencot : Sat Dec 13 00:20:02 CET 2008
 *  Creation
 */

/*
 gcc -g -I. -I.. -I../excp -o ring_test ../ring.c ring_test.c
 */
#include <stdio.h>
#include <errno.h>
#include <assert.h>

#include "ring.h"

int f_ring_test(void) {
   int i, j;
   struct s_ring_test {
      unsigned long stuff;
      struct s_ring link;
      int no;
      unsigned long end;
   } elem[4], *node = 0, *list = 0, out;

   assert ( (list = m_ring_link(list, link, &elem[0])) );
   assert ( !(list = m_ring_link(list, link, &elem[0])) && errno == EEXIST); errno = 0;

   list = 0;
   for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
      assert(list = m_ring_link(list, link, &elem[i]) );
      assert( m_ring_selftest(list, link) );
   }
   
   for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
      assert( m_ring_is_in(&elem[i], list, link) );
      assert( m_ring_is_in(m_ring_next(&elem[i], link), list, link) );
   }
   j = 0;
   m_ring_do(list, node) {
      for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
         if ( &elem[i] == node ) break;
      }
      assert( i < sizeof(elem)/sizeof(elem[0]) );
      ++j;
   } m_ring_done(list, node, link);
   assert( j == sizeof(elem)/sizeof(elem[0]) );

   j = 0;
   for ( node = list; node; node = m_ring_list(list, node, link) ) {
      for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
         if ( &elem[i] == node ) break;
      }
      assert( i < sizeof(elem)/sizeof(elem[0]) );
      ++j;
   }
   assert( j == sizeof(elem)/sizeof(elem[0]) );

   assert( !m_ring_is_in(&out, list, link) );
   assert( !m_ring_is_in(&out, (struct s_ring_test *)0, link) && errno == 0); errno = 0;
   assert( !m_ring_is_in((struct s_ring_test *)0, list, link) && errno == EFAULT); errno = 0;

   assert( (list = m_ring_unlink(&elem[3], link)) && errno == 0); 
   assert ( m_ring_is_in(&elem[0], list, link) );
   assert ( m_ring_is_in(&elem[1], list, link) );
   assert ( m_ring_is_in(&elem[2], list, link) );
   assert (!m_ring_is_in(&elem[3], list, link) );
   assert( (list = m_ring_unlink(&elem[1], link)) && errno == 0); 
   assert ( m_ring_is_in(&elem[0], list, link) );
   assert (!m_ring_is_in(&elem[1], list, link) );
   assert ( m_ring_is_in(&elem[2], list, link) );
   assert (!m_ring_is_in(&elem[3], list, link) );
   assert (list = m_ring_link(list, link, &elem[3]) );
   assert ( m_ring_is_in(&elem[0], list, link) );
   assert (!m_ring_is_in(&elem[1], list, link) );
   assert ( m_ring_is_in(&elem[2], list, link) );
   assert ( m_ring_is_in(&elem[3], list, link) );
   assert( (list = m_ring_unlink(&elem[0], link)) && errno == 0); 
   assert (!m_ring_is_in(&elem[0], list, link) );
   assert (!m_ring_is_in(&elem[1], list, link) );
   assert ( m_ring_is_in(&elem[3], list, link) );
   assert ( m_ring_is_in(&elem[2], list, link) );
   assert( (list = m_ring_unlink(&elem[2], link)) && errno == 0); 
   assert (!m_ring_is_in(&elem[0], list, link) );
   assert (!m_ring_is_in(&elem[1], list, link) );
   assert (!m_ring_is_in(&elem[2], list, link) );
   assert ( m_ring_is_in(&elem[3], list, link) );
   assert( !(list = m_ring_unlink(&elem[3], link)) && errno == 0); 
   assert (!m_ring_is_in(&elem[0], list, link) );
   assert (!m_ring_is_in(&elem[1], list, link) );
   assert (!m_ring_is_in(&elem[2], list, link) );
   assert (!m_ring_is_in(&elem[3], list, link) );
}
int f_nring_test(void) {
   int i, j;
   struct s_nring_test {
      unsigned long stuff;
      struct s_nring link;
      int no;
      unsigned long end;
   } elem[4], *node = 0, *list = 0, out;
   char *names[4] = { "node 0",  "node 1", "node 2", "node 3", };

   elem[0].link.name = names[0];
   assert ( (list = m_nring_link(list, link, &elem[0])) );
   elem[1].link.name = names[0];
   assert ( !(list = m_nring_link(list, link, &elem[1])) && errno == ENOTUNIQ); errno = 0;

   list = 0;
   assert ( (list = m_nring_link(list, link, &elem[0])) );
   assert ( !(list = m_nring_link(list, link, &elem[0])) && errno == EEXIST); errno = 0;

   list = 0;
   for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
      elem[i].link.name = names[i];
   }
      
   for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
      assert(list = m_nring_link(list, link, &elem[i]) );
      assert( m_nring_selftest(list, link) );
   }
   for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
      assert( m_nring_is_in(&elem[i], list, link) );
      assert( m_nring_is_in(m_nring_next(&elem[i], link), list, link) );
      assert( m_nring_find(list, link, names[i]) == &elem[i] );
   }
   j = 0;
   m_nring_do(list, node) {
      for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
         if ( &elem[i] == node ) break;
      }
      assert( i < sizeof(elem)/sizeof(elem[0]) );
      ++j;
   } m_nring_done(list, node, link);
   assert( j == sizeof(elem)/sizeof(elem[0]) );

   j = 0;
   for ( node = list; node; node = m_nring_list(list, node, link) ) {
      for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
         if ( &elem[i] == node ) break;
      }
      assert( i < sizeof(elem)/sizeof(elem[0]) );
      ++j;
   }
   assert( j == sizeof(elem)/sizeof(elem[0]) );

   assert( !m_nring_is_in(&out, list, link) );
   assert( !m_nring_is_in(&out, (struct s_nring_test *)0, link) && errno == 0); errno = 0;
   assert( !m_nring_is_in((struct s_nring_test *)0, list, link) && errno == EFAULT); errno = 0;

   assert( (list = m_nring_unlink(&elem[3], link)) && errno == 0); 
   assert ( m_nring_is_in(&elem[0], list, link) );
   assert ( m_nring_is_in(&elem[1], list, link) );
   assert ( m_nring_is_in(&elem[2], list, link) );
   assert (!m_nring_is_in(&elem[3], list, link) );
   assert( (list = m_nring_unlink(&elem[1], link)) && errno == 0); 
   assert ( m_nring_is_in(&elem[0], list, link) );
   assert (!m_nring_is_in(&elem[1], list, link) );
   assert ( m_nring_is_in(&elem[2], list, link) );
   assert (!m_nring_is_in(&elem[3], list, link) );
   assert (list = m_nring_link(list, link, &elem[3]) );
   assert ( m_nring_is_in(&elem[0], list, link) );
   assert (!m_nring_is_in(&elem[1], list, link) );
   assert ( m_nring_is_in(&elem[2], list, link) );
   assert ( m_nring_is_in(&elem[3], list, link) );
   assert( (list = m_nring_unlink(&elem[0], link)) && errno == 0); 
   assert (!m_nring_is_in(&elem[0], list, link) );
   assert (!m_nring_is_in(&elem[1], list, link) );
   assert ( m_nring_is_in(&elem[3], list, link) );
   assert ( m_nring_is_in(&elem[2], list, link) );
   assert( (list = m_nring_unlink(&elem[2], link)) && errno == 0); 
   assert (!m_nring_is_in(&elem[0], list, link) );
   assert (!m_nring_is_in(&elem[1], list, link) );
   assert (!m_nring_is_in(&elem[2], list, link) );
   assert ( m_nring_is_in(&elem[3], list, link) );
   assert( !(list = m_nring_unlink(&elem[3], link)) && errno == 0); 
   assert (!m_nring_is_in(&elem[0], list, link) );
   assert (!m_nring_is_in(&elem[1], list, link) );
   assert (!m_nring_is_in(&elem[2], list, link) );
   assert (!m_nring_is_in(&elem[3], list, link) );

}
int f_dring_test(void) {
   int i, j;
   struct s_dring_test {
      unsigned long stuff;
      struct s_dring link;
      int no;
      unsigned long end;
   } elem[4], *node = 0, *list = 0, out;

   assert ( (list = m_dring_link(list, link, &elem[0])) );
   assert ( !(list = m_dring_link(list, link, &elem[0])) && errno == EEXIST); errno = 0;

   list = 0;
   for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
      assert(list = m_dring_link(list, link, &elem[i]) );
      assert( m_dring_selftest(list, link) );
   }
   
   for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
      assert( m_dring_is_in(&elem[i], list, link) );
      assert( m_dring_is_in(m_dring_next(&elem[i], link), list, link) );
      assert( m_dring_is_in(m_dring_prev(&elem[i], link), list, link) );
      assert( m_dring_is_in(m_dring_move(&elem[i], 3, 1, link), list, link) );
      assert( m_dring_is_in(m_dring_move(&elem[i],-3, 1, link), list, link) );
      assert( m_dring_is_in(m_dring_move(&elem[i],-3, 1, link), list, link) );
      assert(!m_dring_is_in(m_dring_move(&elem[i],-4, 1, link), list, link) && errno == ELOOP); errno = 0;
      assert(!m_dring_is_in(m_dring_move(&elem[i], 4, 1, link), list, link) && errno == ELOOP); errno = 0;

//printf("node : %08x, move -1 : %08x, prev : %08x, next : %08x\n", &elem[i], m_dring_move(&elem[i],-1, 1, link), m_dring_prev(&elem[i], link), m_dring_next(&elem[i], link));
//fflush(stdout);
      assert( m_dring_move(&elem[i], 4, 0, link) == &elem[i]);
      assert( m_dring_move(&elem[i],-4, 0, link) == &elem[i]);
      assert( m_dring_move(&elem[i], 0, 1, link) == &elem[i] );
      assert( m_dring_move(&elem[i],-1, 1, link) == m_dring_prev(&elem[i], link) );
      assert( m_dring_move(&elem[i], 1, 1, link) == m_dring_next(&elem[i], link) );
      assert( m_dring_prev(m_dring_next(&elem[i], link), link) == &elem[i] );
      assert( m_dring_next(m_dring_prev(&elem[i], link), link) == &elem[i] );
   }
   j = 0;
   m_dring_do(list, node) {
      for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
         if ( &elem[i] == node ) break;
      }
      assert( i < sizeof(elem)/sizeof(elem[0]) );
      ++j;
   } m_dring_done(list, node, link);
   assert( j == sizeof(elem)/sizeof(elem[0]) );

   j = 0;
   for ( node = list; node; node = m_dring_list(list, node, link) ) {
      for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
         if ( &elem[i] == node ) break;
      }
      assert( i < sizeof(elem)/sizeof(elem[0]) );
      ++j;
   }
   assert( j == sizeof(elem)/sizeof(elem[0]) );

   assert( !m_dring_is_in(&out, list, link) );
   assert( !m_dring_is_in(&out, (struct s_dring_test *)0, link) && errno == 0); errno = 0;
   assert( !m_dring_is_in((struct s_dring_test *)0, list, link) && errno == EFAULT); errno = 0;

   assert( (list = m_dring_unlink(&elem[3], link)) && errno == 0); 
   assert ( m_dring_is_in(&elem[0], list, link) );
   assert ( m_dring_is_in(&elem[1], list, link) );
   assert ( m_dring_is_in(&elem[2], list, link) );
   assert (!m_dring_is_in(&elem[3], list, link) );
   assert( (list = m_dring_unlink(&elem[1], link)) && errno == 0); 
   assert ( m_dring_is_in(&elem[0], list, link) );
   assert (!m_dring_is_in(&elem[1], list, link) );
   assert ( m_dring_is_in(&elem[2], list, link) );
   assert (!m_dring_is_in(&elem[3], list, link) );
   assert (list = m_dring_link(list, link, &elem[3]) );
   assert ( m_dring_is_in(&elem[0], list, link) );
   assert (!m_dring_is_in(&elem[1], list, link) );
   assert ( m_dring_is_in(&elem[2], list, link) );
   assert ( m_dring_is_in(&elem[3], list, link) );
   assert( (list = m_dring_unlink(&elem[0], link)) && errno == 0); 
   assert (!m_dring_is_in(&elem[0], list, link) );
   assert (!m_dring_is_in(&elem[1], list, link) );
   assert ( m_dring_is_in(&elem[3], list, link) );
   assert ( m_dring_is_in(&elem[2], list, link) );
   assert( (list = m_dring_unlink(&elem[2], link)) && errno == 0); 
   assert (!m_dring_is_in(&elem[0], list, link) );
   assert (!m_dring_is_in(&elem[1], list, link) );
   assert (!m_dring_is_in(&elem[2], list, link) );
   assert ( m_dring_is_in(&elem[3], list, link) );
   assert( !(list = m_dring_unlink(&elem[3], link)) && errno == 0); 
   assert (!m_dring_is_in(&elem[0], list, link) );
   assert (!m_dring_is_in(&elem[1], list, link) );
   assert (!m_dring_is_in(&elem[2], list, link) );
   assert (!m_dring_is_in(&elem[3], list, link) );
}
int f_ndring_test(void) {
   int i, j;
   struct s_ndring_test {
      unsigned long stuff;
      struct s_ndring link;
      int no;
      unsigned long end;
   } elem[4], *node = 0, *list = 0, out;
   char *names[4] = { "node 0",  "node 1", "node 2", "node 3", };

   elem[0].link.name = names[0];
   assert ( (list = m_ndring_link(list, link, &elem[0])) );
   elem[1].link.name = names[0];
   assert ( !(list = m_ndring_link(list, link, &elem[1])) && errno == ENOTUNIQ); errno = 0;

   list = 0;
   assert ( (list = m_ndring_link(list, link, &elem[0])) );
   assert ( !(list = m_ndring_link(list, link, &elem[0])) && errno == EEXIST); errno = 0;

   list = 0;
   for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
      elem[i].link.name = names[i];
   }
      
   assert ( (list = m_ndring_link(list, link, &elem[0])) );
   assert ( !(list = m_ndring_link(list, link, &elem[0])) && errno == EEXIST); errno = 0;

   list = 0;
   for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
      assert(list = m_ndring_link(list, link, &elem[i]) );
      assert( m_ndring_selftest(list, link) );
   }
   
   for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
      assert( m_ndring_is_in(&elem[i], list, link) );
      assert( m_ndring_find(list, link, names[i]) == &elem[i] );
      assert( m_ndring_is_in(m_ndring_next(&elem[i], link), list, link) );
      assert( m_ndring_is_in(m_ndring_prev(&elem[i], link), list, link) );
      assert( m_ndring_is_in(m_ndring_move(&elem[i], 3, 1, link), list, link) );
      assert( m_ndring_is_in(m_ndring_move(&elem[i],-3, 1, link), list, link) );
      assert( m_ndring_is_in(m_ndring_move(&elem[i],-3, 1, link), list, link) );
      assert(!m_ndring_is_in(m_ndring_move(&elem[i],-4, 1, link), list, link) && errno == ELOOP); errno = 0;
      assert(!m_ndring_is_in(m_ndring_move(&elem[i], 4, 1, link), list, link) && errno == ELOOP); errno = 0;

//printf("node : %08x, move -1 : %08x, prev : %08x, next : %08x\n", &elem[i], m_ndring_move(&elem[i],-1, 1, link), m_ndring_prev(&elem[i], link), m_ndring_next(&elem[i], link));
//fflush(stdout);
      assert( m_ndring_move(&elem[i], 4, 0, link) == &elem[i]);
      assert( m_ndring_move(&elem[i],-4, 0, link) == &elem[i]);
      assert( m_ndring_move(&elem[i], 0, 1, link) == &elem[i] );
      assert( m_ndring_move(&elem[i],-1, 1, link) == m_ndring_prev(&elem[i], link) );
      assert( m_ndring_move(&elem[i], 1, 1, link) == m_ndring_next(&elem[i], link) );
      assert( m_ndring_prev(m_ndring_next(&elem[i], link), link) == &elem[i] );
      assert( m_ndring_next(m_ndring_prev(&elem[i], link), link) == &elem[i] );
   }
   j = 0;
   m_ndring_do(list, node) {
      for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
         if ( &elem[i] == node ) break;
      }
      assert( i < sizeof(elem)/sizeof(elem[0]) );
      ++j;
   } m_ndring_done(list, node, link);
   assert( j == sizeof(elem)/sizeof(elem[0]) );

   j = 0;
   for ( node = list; node; node = m_ndring_list(list, node, link) ) {
      for ( i = 0; i < sizeof(elem)/sizeof(elem[0]); i++ ) {
         if ( &elem[i] == node ) break;
      }
      assert( i < sizeof(elem)/sizeof(elem[0]) );
      ++j;
   }
   assert( j == sizeof(elem)/sizeof(elem[0]) );

   assert( !m_ndring_is_in(&out, list, link) );
   assert( !m_ndring_is_in(&out, (struct s_ndring_test *)0, link) && errno == 0); errno = 0;
   assert( !m_ndring_is_in((struct s_ndring_test *)0, list, link) && errno == EFAULT); errno = 0;

   assert( (list = m_ndring_unlink(&elem[3], link)) && errno == 0); 
   assert ( m_ndring_is_in(&elem[0], list, link) );
   assert ( m_ndring_is_in(&elem[1], list, link) );
   assert ( m_ndring_is_in(&elem[2], list, link) );
   assert (!m_ndring_is_in(&elem[3], list, link) );
   assert( (list = m_ndring_unlink(&elem[1], link)) && errno == 0); 
   assert ( m_ndring_is_in(&elem[0], list, link) );
   assert (!m_ndring_is_in(&elem[1], list, link) );
   assert ( m_ndring_is_in(&elem[2], list, link) );
   assert (!m_ndring_is_in(&elem[3], list, link) );
   assert (list = m_ndring_link(list, link, &elem[3]) );
   assert ( m_ndring_is_in(&elem[0], list, link) );
   assert (!m_ndring_is_in(&elem[1], list, link) );
   assert ( m_ndring_is_in(&elem[2], list, link) );
   assert ( m_ndring_is_in(&elem[3], list, link) );
   assert( (list = m_ndring_unlink(&elem[0], link)) && errno == 0); 
   assert (!m_ndring_is_in(&elem[0], list, link) );
   assert (!m_ndring_is_in(&elem[1], list, link) );
   assert ( m_ndring_is_in(&elem[3], list, link) );
   assert ( m_ndring_is_in(&elem[2], list, link) );
   assert( (list = m_ndring_unlink(&elem[2], link)) && errno == 0); 
   assert (!m_ndring_is_in(&elem[0], list, link) );
   assert (!m_ndring_is_in(&elem[1], list, link) );
   assert (!m_ndring_is_in(&elem[2], list, link) );
   assert ( m_ndring_is_in(&elem[3], list, link) );
   assert( !(list = m_ndring_unlink(&elem[3], link)) && errno == 0); 
   assert (!m_ndring_is_in(&elem[0], list, link) );
   assert (!m_ndring_is_in(&elem[1], list, link) );
   assert (!m_ndring_is_in(&elem[2], list, link) );
   assert (!m_ndring_is_in(&elem[3], list, link) );
}

int main(int argc, char *argv[]) {
   f_ring_check_opt(1);
   f_ring_test();
   f_nring_test();
   f_dring_test();
   f_ndring_test();
   return 0;
}

